import 'dart:convert';

import 'package:flutter/cupertino.dart';
import 'package:flutter_bloc/flutter_bloc.dart';

import 'constants.dart';
import 'one_to_one_event.dart';
import 'one_to_one_state.dart';
import 'one_to_one_usecase_service.dart';
import 'socket_state.dart';

/// bloc class for video call one to one tranmedika or business logic one to one call ///
class OneToOneBloc extends Bloc<OneToOneUiUIEvent, OneToOneUiState>{
  OneToOneBloc(
      {required OneToOneUseCaseService oneToOneUseCaseService}):
        _oneToOneUseCaseService = oneToOneUseCaseService, super(const OneToOneUiState(
        socketState: SocketState(socketConnectionState: SocketConnectionState.idle)
  )) {
    /// register event ///
    on<OnConnectToSocket>(
        _onConnectToSocket
    );

    on<OnSocketOpen>(
        _onSocketOpen
    );

    on<OnSocketMessage>(
        _onSocketMessage
    );

    on<OnSocketClose>(
        _onSocketClose
    );

    on<OnRegister>(
        _onRegisterToSocket
    );

    on<OnLocalStream>(
        _onLocalStream
    );

    on<OnAddRemoteStream>(
        _onAddRemoteStream
    );

    on<OnRemoveRemoteStream>(
        _onRemoveRemoteStream
    );

    on<OnCloseSession>(
      _onCloseSession
    );

    on<OnCallAccepted>(
      _onCallAccepted
    );

    on<OnCallRejected>(
      _onCallRejected
    );

    on<OnStopCall>(
      _onStopCall
    );

    on<OnMute>(
      _onMute
    );

    on<OnSwitchCamera>(
        _onSwitchCamera
    );

    on<OnEnableVideo>(
        _onEnableVideo
    );

    on<OnScreenShare>(
        _onScreenShare
    );

    on<OnSwapScreen>(
        _onSwapScreen
    );
  }


  final OneToOneUseCaseService _oneToOneUseCaseService;

  _onCloseSession(OnCloseSession event, Emitter<OneToOneUiState> emit) async{
    await _oneToOneUseCaseService.closeSession();
  }

  /// event ketika register ke web socket transmedika ///
  _onRegisterToSocket(OnRegister event, Emitter<OneToOneUiState> emit) async{
    emit(state.copyWith(callState: CallState.registering));
    await _oneToOneUseCaseService.register(from: state.from!, to: state.to!, isCalling: state.isCalling!);
  }

  /// event ketika web socket connected ///
  _onSocketOpen(OnSocketOpen event, Emitter<OneToOneUiState> emit) async{
    emit(state.copyWith(
        socketState: const SocketState(
            socketConnectionState: SocketConnectionState.open
        )
    ));
    add(OnRegister());
  }

  /// event ketika menerima local stream ///
  _onLocalStream(OnLocalStream event, Emitter<OneToOneUiState> emit) async{
    emit(state.copyWith(localStream: event.mediaStream));
  }

  /// event ketika menerima remote stream ///
  _onAddRemoteStream(OnAddRemoteStream event, Emitter<OneToOneUiState> emit) async{
    emit(state.copyWith(remoteStream: event.mediaStream));
  }

  /// event ketika hapus remote stream ///
  _onRemoveRemoteStream(OnRemoveRemoteStream event, Emitter<OneToOneUiState> emit) async{
    emit(state.copyWith(removeRemoteStream: event.mediaStream));
  }

  /// event ketika user reject video call ///
  _onCallRejected(OnCallRejected event, Emitter<OneToOneUiState> emit) async {
    _oneToOneUseCaseService.reject(from: state.registrationSessionFrom!, to: state.registrationSessionTo!);
    emit(state.copyWith(callState: CallState.callRejectedByUser));
    await _oneToOneUseCaseService.close();
  }

  /// event ketika enable disable mic ///
  _onMute(OnMute event, Emitter<OneToOneUiState> emit) async {
    if(state.mute == true){
      emit(state.copyWith(mute: false));
    }else{
      emit(state.copyWith(mute: true));
    }
    _oneToOneUseCaseService.muteMic();
  }

  /// event ketika swap screen ///
  _onSwapScreen(OnSwapScreen event, Emitter<OneToOneUiState> emit) async {
    if(state.swapScreen == true){
      emit(state.copyWith(swapScreen: false));
    }else{
      emit(state.copyWith(swapScreen: true));
    }
  }

  /// event ketika Screen Share ///
  _onScreenShare(OnScreenShare event, Emitter<OneToOneUiState> emit) async{
    emit(state.copyWith(videoSource: VideoSource.screen));

    await _oneToOneUseCaseService.selectScreenSource(
        videoSource: state.videoSource!, onLocalStream: (stream){
        add(OnLocalStream(mediaStream: stream));
    });
  }

  /// event ketika enable disable video ///
  _onEnableVideo(OnEnableVideo event, Emitter<OneToOneUiState> emit) async {
    if(state.enableVideo == true){
      emit(state.copyWith(enableVideo: false));
    }else{
      emit(state.copyWith(enableVideo: true));
    }
    _oneToOneUseCaseService.enableVideo();
  }

  /// event ketika user switch camera front or rear ///
  _onSwitchCamera(OnSwitchCamera event, Emitter<OneToOneUiState> emit) async {
    if(state.frontCamera == true){
      emit(state.copyWith(frontCamera: false));
    }else{
      emit(state.copyWith(frontCamera: true));
    }

    _oneToOneUseCaseService.switchCamera(videoSource: state.videoSource!, onLocalStream: (stream){
      add(OnLocalStream(mediaStream: stream));
    });

    if(state.videoSource == VideoSource.screen){
      emit(state.copyWith(videoSource: VideoSource.camera));
    }
  }

  /// event ketika user stop video call ///
  _onStopCall(OnStopCall event, Emitter<OneToOneUiState> emit) async {
    await _oneToOneUseCaseService.closeSession();
    await _oneToOneUseCaseService.close();
  }

  /// event ketika user menerima video call ///
  _onCallAccepted(OnCallAccepted event, Emitter<OneToOneUiState> emit) async {
    emit(state.copyWith(callState: CallState.callAcceptedByUser));
    await _createLocalPeerConnection();
    _createOffer(
        state.registrationSessionFrom!,
        state.registrationSessionTo!,
        state.isCalling!,
        {
          'consultationId': state.consultationId,
          'callerUUID': state.mySelf!['uuid'],
          'calleeUUID': state.other!['uuid']
        }
    );
  }

  /// Event receive message from web socket ///
  _onSocketMessage(OnSocketMessage event, Emitter<OneToOneUiState> emit) async{
    Map<String, dynamic> valueMap = json.decode(event.message);

    emit(state.copyWith(
        socketState: state.socketState?.copyWith(
          message: event.message,
          messageModel: valueMap
        )
    ));

    switch (valueMap['id']) {
      case Constants.registerResponse:
        Map<String, dynamic> registrationPayload = valueMap['registrationPayload'];
        if (valueMap['response'] == Constants.accepted) {
          emit(state.copyWith(
              callState: CallState.registerAccepted,
              registrationSessionFrom: registrationPayload
          ));
          if(state.isCalling == false) {
            await _oneToOneUseCaseService.notifyCaller(from: registrationPayload, to: state.to!);
          }else{
            await _oneToOneUseCaseService.callingMobileDevice(from: state.from!, to: state.to!, fcm: state.fcm!);
          }
        } else {
          emit(state.copyWith(
            callState: CallState.registerRejected,
            registrationSessionFrom: registrationPayload
          ));
        }
        break;
      case Constants.notifyCallerActive:
        emit(state.copyWith(
            callState: CallState.notifyMe,
            registrationSessionTo: valueMap['from']
        ));

        if(state.isCalling == true){
          await _createLocalPeerConnection();
          _createOffer(
            state.registrationSessionFrom!,
            state.registrationSessionTo!,
            state.isCalling!,
              {
                'consultationId': state.consultationId!,
                'callerUUID': state.mySelf!['uuid'],
                'calleeUUID': state.other!['uuid']
              }
          );
        }
        break;
      case Constants.callingResponse:
        if (valueMap['response'] == Constants.accepted) {
          emit(state.copyWith(callState: CallState.callingAccepted));
        } else {
          emit(state.copyWith(callState: CallState.callingRejected));
        }
        break;
      case Constants.notifyCallerNotActive:
        emit(state.copyWith(
            callState: CallState.missedCall,
            registrationSessionFrom: valueMap['from']
        ));
        break;
      case Constants.callResponse:
        if (valueMap['response'] == Constants.accepted) {
          emit(state.copyWith(callState: CallState.callAccepted));
          _oneToOneUseCaseService.receiveVideoAnswer(valueMap['sdpAnswer']);
        } else {
          emit(state.copyWith(callState: CallState.callRejected));
        }
        break;
      case Constants.incomingCall:
        emit(state.copyWith(
          callState: CallState.incomingCall,
          registrationSessionTo: valueMap['from']
        ));
        break;
      case Constants.startCommunication:
        emit(state.copyWith(callState: CallState.startCommunication));
        _oneToOneUseCaseService.receiveVideoAnswer(valueMap['sdpAnswer']);
        break;
      case Constants.iceCandidate:
        emit(state.copyWith(callState: CallState.iceCandidate));
        _oneToOneUseCaseService.iceCandidate(valueMap['candidate']);
        break;
      case Constants.stopCommunication:
        emit(state.copyWith(callState: CallState.stopCommunication));
        break;
      default:
        emit(state.copyWith(callState: CallState.unknown));
        debugPrint(Constants.unknown);
    }
  }

  /// create peer connection ///
  _createLocalPeerConnection() async{
    /// closure from createLocalPeerConnection return onLocalStream, onRemoteStream, onRemoveRemoteStream
    await _oneToOneUseCaseService.createLocalPeerConnection(
        onLocalStream: (stream) {
          add(OnLocalStream(mediaStream: stream));
        },
        onAddRemoteStream: (stream) {
          add(OnAddRemoteStream(mediaStream: stream));
        },
        onRemoveRemoteStream: (stream){
          add(OnRemoveRemoteStream(mediaStream: stream));
        }
    );
  }

  /// create offer ///
  _createOffer(Map<String, dynamic> from, Map<String, dynamic> to, bool isCalling, Map<String, dynamic>? consultation) async{
    _oneToOneUseCaseService.createOffer(from: from, to: to, isCalling: isCalling, consultation: consultation);
  }

  /// event on close web socket ///
  _onSocketClose(OnSocketClose event, Emitter<OneToOneUiState> emit) async{
    if(event.from){
      await _oneToOneUseCaseService.close();
    }
    emit(state.copyWith(
        socketState: SocketState(
            socketConnectionState: SocketConnectionState.close,
            exception: event.ex
        )
    ));
  }

  /// event connect web socket ///
  _onConnectToSocket(OnConnectToSocket event, Emitter<OneToOneUiState> emit) async{
    /// emit or set to state ///

    Map<String,dynamic>? fcm;
    if(event.fcmTokens!=null) {
      final data = {
        'body': '${event.mySelf['email']} Menelepon',
        'title': 'PANGGILAN',
        'app': 'VIDEO_CALL',
        'status': 'CALLING',
        'consultation_id': event.consultationId,
        'me': const JsonEncoder().convert(event.mySelf),
        'friend': const JsonEncoder().convert(event.mySelf)
      };

      fcm = {
        'to': event.fcmTokens!,
        'data': data
      };
    }

    emit(state.copyWith(
      from: event.from,
      to: event.to,
      isCalling: event.isCalling,
      socketState: const SocketState(
        socketConnectionState: SocketConnectionState.connecting
      ),
      mySelf: event.mySelf,
      other: event.other,
      consultationId: event.consultationId,
      fcm: fcm
    ));
    try{
      /// closure from method connect, return onopen, onmessage, onclose ///
      await _oneToOneUseCaseService.connect(url: event.host,
              onOpen: () => add(OnSocketOpen()),
              onMessage: (message) => add(OnSocketMessage(message: message)),
              onClose: (code, reason) => {
                //add(OnSocketClose(ex: reason!, from: false))
              }
      );

    }on Exception catch(e){
      emit(state.copyWith(
        socketState: SocketState(
            socketConnectionState: SocketConnectionState.error,
            exception: e.toString())
      ));
    }
  }

  @override
  Future<void> close() async {
    _oneToOneUseCaseService.close();
    return super.close();
  }


}