import 'package:cached_network_image/cached_network_image.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:flutter_webrtc/flutter_webrtc.dart';

import 'one_to_one_bloc.dart';
import 'one_to_one_event.dart';
import 'one_to_one_state.dart';
import 'one_to_one_usecase.dart';
import 'socket_state.dart';

/// one to one screen ///
class OneToOneScreen extends StatelessWidget{
  const OneToOneScreen({
    Key? key,
    required this.consultationId,
    required this.mySelf,
    required this.other,
    this.fcmTokens,
    this.onMessage
  }): super(key: key);

  final String consultationId;
  final Map<String, dynamic> mySelf;
  final Map<String, dynamic> other;
  final List<String>? fcmTokens;
  final Function(String message)? onMessage;

  @override
  Widget build(BuildContext context) => RepositoryProvider.value(
      value: OneToOneUseCase(),
      child: BlocProvider<OneToOneBloc>(
          create: (context) =>
              OneToOneBloc(
                  oneToOneUseCaseService: context.read<OneToOneUseCase>()
              ),
          child: OneToOnePage(consultationId: consultationId, mySelf: mySelf, other: other, fcmTokens: fcmTokens)
      )
  );
}

class OneToOnePage extends StatefulWidget {
  const OneToOnePage({
    Key? key,
    required this.consultationId,
    required this.mySelf,
    required this.other,
    this.fcmTokens,
    this.onMessage
  }): super(key: key);
  final String consultationId;
  final Map<String, dynamic> mySelf;
  final Map<String, dynamic> other;
  final List<String>? fcmTokens;
  final Function(String message)? onMessage;

  @override
  State<OneToOnePage> createState() => _OneToOnePageState();
}

class _OneToOnePageState extends State<OneToOnePage> {

  final RTCVideoRenderer _localRenderer = RTCVideoRenderer();
  final RTCVideoRenderer _remoteRenderer = RTCVideoRenderer();

  /// on user trying to connet web socket ///
  _onConnectToSocket({
    required String url,
    required String from,
    required String to,
    required bool isCalling,
    required Map<String, dynamic> mySelf,
    required Map<String, dynamic> other,
    required String consultationId,
    List<String>? fcmTokens
  }) async {
    context
        .read<OneToOneBloc>()
        .add(OnConnectToSocket(
          host: url,
          from: from,
          to: to,
          isCalling: isCalling,
          mySelf: mySelf,
          other: other,
          consultationId: consultationId,
          fcmTokens: fcmTokens
      )
    );
  }

  /*_onCloseSocket() async {
    context
        .read<OneToOneBloc>()
        .add(OnSocketClose(ex: "Close Socket from UI", from: true));
  }*/

  /// close session from user ///
  _onCloseSession() async{
    context
        .read<OneToOneBloc>()
        .add(OnCloseSession());
  }

  /// call accepted by user ///
  _onCallAcceptedByUser() async{
    context
        .read<OneToOneBloc>()
        .add(OnCallAccepted());
  }

  /// call rejected by user ///
  _onCallRejectedByUser() async{
    context
        .read<OneToOneBloc>()
        .add(OnCallRejected());
  }

  /// stop video call by user ///
  _onStopCall() async{
    context
        .read<OneToOneBloc>()
        .add(OnStopCall());
  }

  /// mute video call from user ///
  _onMute() async{
    context
        .read<OneToOneBloc>()
        .add(OnMute());
  }

  /// mute video call from user ///
  _onSwapScreen() async{
    context
        .read<OneToOneBloc>()
        .add(OnSwapScreen());
  }

  /// swicth camera front or rear from user ///
  _onSwitchCamera() async{
    context
        .read<OneToOneBloc>()
        .add(OnSwitchCamera());
  }

  /// enable disable video ///
  _onEnableVideo() async{
    context
        .read<OneToOneBloc>()
        .add(OnEnableVideo());
  }

  /// Screen Share ///
  _onScreenShare() async{
    context
        .read<OneToOneBloc>()
        .add(OnScreenShare());
  }

  /// initialize local and remote renderer ///
  _initRenderers() async {
    await _localRenderer.initialize();
    await _remoteRenderer.initialize();
  }

  _stop(){
    _onCloseSession();
    _localRenderer.srcObject = null;
    _remoteRenderer.srcObject = null;
  }

  /// dispose renderer ///
  _disposeRenderers() {
    _stop();
    _localRenderer.dispose();
    _remoteRenderer.dispose();
  }

  @override
  void initState() {
    _initRenderers();
    _onConnectToSocket(
      url: "wss://rtc.mybeam.me/one2one",
      from: widget.mySelf['email'],
      to: widget.other['email'],
      isCalling: widget.fcmTokens != null ? true : false,
      mySelf: widget.mySelf,
      other: widget.other,
      consultationId: widget.consultationId,
      fcmTokens: widget.fcmTokens
    );
    super.initState();
  }

  @override
  void deactivate() {
    _disposeRenderers();
    //_onCloseSocket();
    super.deactivate();
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
        body: BlocConsumer<OneToOneBloc, OneToOneUiState>(
            listenWhen: (previous, current) {
              /// build change when old state != current state
              return previous!=current;
            },
            listener: (context, state) {
              /// logic web socket state ///
              debugPrint('State: ${state.socketState?.socketConnectionState.toString()}');
              switch(state.socketState?.socketConnectionState){
                case SocketConnectionState.connecting :
                  break;
                case SocketConnectionState.open :
                  debugPrint('Message: ${state.socketState?.message}');
                  break;
                case SocketConnectionState.close :
                case SocketConnectionState.error :
                  debugPrint('Error: ${state.socketState?.exception}');
                  break;
                default:
              }

              /// on missed call from caller ///
              if(state.callState == CallState.missedCall){
                widget.onMessage?.call("Missed call from ${state.other?['name']}");
              }
            },
            builder: (context, state) {
              String? socketStateStr = "";
              Color socketColor = Colors.white;

              /// logic transmedika call state ///
              switch(state.socketState?.socketConnectionState){
                case SocketConnectionState.connecting:
                  socketStateStr = "Connecting";
                  socketColor = Colors.yellow;
                  break;
                case SocketConnectionState.open:
                  socketStateStr = "Connected";
                  socketColor = Colors.green;
                  break;
                case SocketConnectionState.idle:
                  socketStateStr = "Idle";
                  socketColor = Colors.blue;
                  break;
                case SocketConnectionState.error:
                case SocketConnectionState.close:
                  socketStateStr = state.socketState?.exception;
                  socketColor = Colors.red;
                  break;
                default:
              }

              String? callStateStr = "";
              Color callColor = Colors.white;
              switch(state.callState){
                case CallState.registering:
                  callStateStr = "Registering";
                  callColor = Colors.yellow;
                  break;
                case CallState.registerAccepted:
                  callStateStr = "Register Accepted and ${state.isCalling == false ? 'Notify Caller ${state.other?['name']}' : 'Calling ${state.other?['name']}'} ";
                  callColor = Colors.green;
                  break;
                case CallState.registerRejected:
                  callStateStr = "Register Rejected";
                  callColor = Colors.red;
                  break;
                case CallState.callingAccepted:
                  callStateStr = 'Calling ${state.other?['name']}';
                  callColor = Colors.green;
                  break;
                case CallState.callingRejected:
                  callStateStr = "Call Rejected";
                  callColor = Colors.red;
                  _onCloseSession();
                  break;
                case CallState.callAccepted:
                  callStateStr = "Call Accepted";
                  callColor = Colors.green;
                  break;
                case CallState.callRejected:
                  callStateStr = "Call Rejected";
                  callColor = Colors.red;
                  _onCloseSession();
                  break;
                case CallState.incomingCall:
                  Map<String, dynamic> from = state.socketState?.messageModel?['from'];
                  callStateStr = "Incoming Call From: ${from['name']}";
                  callColor = Colors.green;
                  break;
                case CallState.startCommunication:
                  callStateStr = "Start Communication";
                  callColor = Colors.green;
                  break;
                case CallState.iceCandidate:
                  callStateStr = "Ice Candidate/On Call";
                  callColor = Colors.green;
                  break;
                case CallState.notifyMe:
                  callStateStr = state.isCalling == true ? 'Ringing' : 'Calling ${state.other?['name']}';
                  callColor = Colors.green;
                  break;
                case CallState.stopCommunication:
                  callStateStr = "Stop Communication";
                  callColor = Colors.green;
                  _stop();
                  if (Navigator.canPop(context)) {
                    Navigator.pop(context);
                  } else {
                    SystemNavigator.pop();
                  }
                  break;
                case CallState.missedCall:
                  _stop();
                  if (Navigator.canPop(context)) {
                    Navigator.pop(context);
                  } else {
                    SystemNavigator.pop();
                  }
                  break;
                default:
              }

              /// display video remote if remote stream available ///
              if(state.remoteStream!=null){
                if(state.swapScreen == true){
                  _remoteRenderer.srcObject = state.localStream;
                }else{
                  _remoteRenderer.srcObject = state.remoteStream;
                }
              }

              if(state.removeRemoteStream!=null){
                _remoteRenderer.srcObject = null;
              }

              /// display video local if local stream available ///
              if(state.localStream!=null) {
                if(state.swapScreen == true){
                  _localRenderer.srcObject = state.remoteStream;
                }else{
                  _localRenderer.srcObject = state.localStream;
                }
              }

              return OrientationBuilder(builder: (context, orientation) {
                final btnBlueStyle = ButtonStyle(
                  backgroundColor: MaterialStateProperty.all(Colors.blue),
                  padding: MaterialStateProperty.all(const EdgeInsets.all(20)),
                  shape: MaterialStateProperty.all(const CircleBorder()),
                );

                final btnRedStyle = ButtonStyle(
                  backgroundColor: MaterialStateProperty.all(Colors.red),
                  padding: MaterialStateProperty.all(const EdgeInsets.all(20)),
                  shape: MaterialStateProperty.all(const CircleBorder()),
                );

                final btnGreenStyle = ButtonStyle(
                  backgroundColor: MaterialStateProperty.all(Colors.green),
                  padding: MaterialStateProperty.all(const EdgeInsets.all(20)),
                  shape: MaterialStateProperty.all(const CircleBorder()),
                );
                /// view remote renderer ///
                return Stack(children: <Widget>[
                  Positioned(
                      left: 0.0,
                      right: 0.0,
                      top: 0.0,
                      bottom: 0.0,
                      child: Container(
                        margin: const EdgeInsets.fromLTRB(0.0, 0.0, 0.0, 0.0),
                        width: MediaQuery.of(context).size.width,
                        height: MediaQuery.of(context).size.height,
                        decoration: const BoxDecoration(color: Colors.black54),
                        child: RTCVideoView(_remoteRenderer),
                      )
                  ),
                  /// view local renderer ///
                  Positioned(
                    left: 20.0,
                    top: 40.0,
                    child: Visibility(
                      visible: state.localStream!=null,
                      child:
                      GestureDetector(
                        onDoubleTap: (){
                          _onSwapScreen();
                        },
                        child: Container(
                          width: orientation == Orientation.portrait ? 110.0 : 140.0,
                          height: orientation == Orientation.portrait ? 140.0 : 110.0,
                          decoration: const BoxDecoration(color: Colors.black54),
                          child: RTCVideoView(_localRenderer),
                        ),
                      )
                    )
                  ),
                  /// view video settings ///
                  Align(
                    alignment: Alignment.topRight,
                    child: Visibility(
                      visible: state.localStream!=null,
                      child: Padding(
                          padding: const EdgeInsets.all(16.0),
                          child: Column(
                            crossAxisAlignment: CrossAxisAlignment.start,
                            mainAxisSize: MainAxisSize.min,
                            children: [
                              const SizedBox(height: 20),
                              ElevatedButton(
                                onPressed: () {
                                  _onSwitchCamera();
                                },
                                style: btnBlueStyle,
                                child: state.frontCamera == false ? const Icon(Icons.camera_front, color: Colors.white) : const Icon(Icons.camera_rear, color: Colors.white),
                              ),
                              const SizedBox(height: 10),
                              ElevatedButton(
                                onPressed: () {
                                  _onMute();
                                },
                                style: btnBlueStyle,
                                child: state.mute == false ? const Icon(Icons.mic_off, color: Colors.white) : const Icon(Icons.mic, color: Colors.white),
                              ),
                              const SizedBox(height: 10),
                              ElevatedButton(
                                onPressed: () {
                                  _onEnableVideo();
                                },
                                style: btnBlueStyle,
                                child: state.enableVideo == true ? const Icon(Icons.videocam_off, color: Colors.white) : const Icon(Icons.videocam, color: Colors.white),
                              ),
                              const Visibility(
                                visible: kIsWeb ? true : false,
                                child: SizedBox(height: 10),
                              ),
                              Visibility(
                                visible: kIsWeb ? true : false,
                                child: ElevatedButton(
                                  onPressed: () {
                                    _onScreenShare();
                                  },
                                  style: btnBlueStyle,
                                  child: state.videoSource == VideoSource.camera ? const Icon(Icons.screen_share, color: Colors.white) : const Icon(Icons.stop_screen_share, color: Colors.white),
                                )
                              )
                            ],
                          )
                      )
                    )
                  ),

                  /// view accept and reject button ///
                  Align(
                    alignment: orientation == Orientation.portrait ? Alignment.bottomCenter: Alignment.bottomRight,
                    child: Padding(
                      padding: orientation == Orientation.portrait ? const EdgeInsets.only(bottom: 80) : const EdgeInsets.only(bottom: 16, right: 16),
                      child: Row(
                        crossAxisAlignment: CrossAxisAlignment.start,
                        mainAxisSize: MainAxisSize.min,
                        children: [
                          Visibility(
                            visible: state.callState == CallState.incomingCall ||
                                state.callState == CallState.startCommunication ||
                                state.callState == CallState.iceCandidate,
                            child: ElevatedButton(
                              onPressed: () {
                                if(state.callState == CallState.incomingCall){
                                  _onCallRejectedByUser();
                                }else{
                                  _onStopCall();
                                }
                                _stop();
                                if (Navigator.canPop(context)) {
                                  Navigator.pop(context);
                                } else {
                                  SystemNavigator.pop();
                                }
                              },
                              style: btnRedStyle,
                              child: const Icon(Icons.call_end, color: Colors.white),
                            ),
                          ),
                          Visibility(
                            visible: state.callState == CallState.incomingCall,
                            child: orientation == Orientation.portrait ? const SizedBox(width: 50) : const SizedBox(width: 30),
                          ),
                          Visibility(
                            visible: state.callState == CallState.incomingCall,
                            child: ElevatedButton(
                              onPressed: () {
                                _onCallAcceptedByUser();
                              },
                              style: btnGreenStyle,
                              child: const Icon(Icons.call, color: Colors.white),
                            ),
                          )
                        ],
                      )
                    )
                  ),
                  Align(
                    alignment: Alignment.topCenter,
                    child: Visibility(
                      visible: state.callState == null || state.callState == CallState.incomingCall ||
                          state.callState == CallState.registering || state.callState == CallState.registerAccepted ||
                          state.callState == CallState.callingAccepted || state.callState == CallState.notifyMe,
                      child: Padding(
                        padding: const EdgeInsets.only(top:64, right: 32, left: 32),
                        child: Column(
                          crossAxisAlignment: CrossAxisAlignment.center,
                          mainAxisSize: MainAxisSize.min,
                          children: [
                            ClipRRect(
                              borderRadius: BorderRadius.circular(120),
                              child: CachedNetworkImage(
                                width: 190,
                                height: 190,
                                imageUrl: state.other?['profilePicture'] ?? '',
                                imageBuilder: (context, imageProvider) => Container(
                                  decoration: BoxDecoration(
                                    image: DecorationImage(
                                        image: imageProvider,
                                        fit: BoxFit.cover),
                                  ),
                                ),
                                placeholder: (context, url) => const Icon(
                                  Icons.account_circle_outlined,
                                  color: Colors.white,
                                  size: 190,
                                ),
                                errorWidget: (context, url, error) => const Icon(
                                  Icons.account_circle_outlined,
                                  color: Colors.red,
                                  size: 190,
                                ),
                              ),
                            ),
                            const SizedBox(height: 30,),
                            Container(
                              padding: const EdgeInsets.only(right: 80.0, left: 80),
                              child: Text(state.other?['name'] ?? '-',
                                overflow: TextOverflow.ellipsis,
                                maxLines: 1,
                                style: const TextStyle(
                                    color: Colors.white,
                                    fontSize: 20
                                ),
                              ),
                            ),
                          ],
                        )
                      ),
                    )
                  ),

                  /// view check state video call ///
                  Align(
                    alignment: Alignment.bottomLeft,
                    child: Padding(
                      padding: const EdgeInsets.all(16.0),
                      child: Column(
                        crossAxisAlignment: CrossAxisAlignment.start,
                        mainAxisSize: MainAxisSize.min,
                        children: [
                          RichText(
                            text: TextSpan(
                              text: 'Socket State: ',
                              style: const TextStyle(color: Colors.white),
                              children: <TextSpan>[
                                TextSpan(text: socketStateStr!,
                                    style: TextStyle(
                                        fontWeight: FontWeight.bold,
                                        color: socketColor
                                    )),
                              ],
                            ),
                          ),
                          RichText(
                            text: TextSpan(
                              text: 'Call State: ',
                              style: const TextStyle(color: Colors.white),
                              children: <TextSpan>[
                                TextSpan(text: callStateStr!,
                                    style: TextStyle(
                                        fontWeight: FontWeight.bold,
                                        color: callColor
                                    )),
                              ],
                            ),
                          ),
                        ],
                      )
                    )
                  ),
                ]);
              });
            }
        )
    );
  }
}