import 'dart:convert';

import 'package:flutter/foundation.dart';
import 'package:flutter_webrtc/flutter_webrtc.dart';
import 'package:transmedika_one_to_one_call/webrtc/one_to_one_state.dart';
import 'package:transmedika_one_to_one_call/webrtc/socket_usecase.dart';

import 'constants.dart';
import 'one_to_one_usecase_service.dart';

/// Implement Use Case one to one video call ///
class OneToOneUseCase extends SocketUseCase implements OneToOneUseCaseService{
  late RTCPeerConnection? _peerConnection;
  late MediaStream? _localStream;
  final List<RTCIceCandidate> remoteCandidates = <RTCIceCandidate>[];
  final List<RTCRtpSender> _senders = <RTCRtpSender>[];
  final List<MediaStream> _remoteStreams = <MediaStream>[];

  /// Register to web socket transmedika ///
  @override
  Future<void> register({required String from, required String to, required bool isCalling}) async {
    final param = {
      'id': Constants.register,
      'from': from,
      'to': to,
      'isCalling': isCalling
    };
    await getSocket().send(const JsonEncoder().convert(param));
  }

  /// Create peer connection ///
  @override
  Future<void> createLocalPeerConnection({
    required Function(MediaStream stream) onLocalStream,
    required Function(MediaStream stream) onAddRemoteStream,
    required Function(MediaStream stream)? onRemoveRemoteStream
  }) async {
    _localStream = await createStream(); /// create stream ///
    onLocalStream.call(_localStream!);

    /// create peer connection and set webrtc configuration ///
    _peerConnection = await createPeerConnection({
      ...Constants.iceServers,
      ...Constants.rtcConfig
    }, Constants.loopbackConstraints);

    /// on receive ice candidate from local ///
    _peerConnection?.onIceCandidate = (candidate) {
      onIceCandidate(candidate);
    };

    switch (Constants.sdpSemantics) {
      case Constants.planB:
        _peerConnection?.onAddStream = (MediaStream stream) {
          onAddRemoteStream.call(stream);
          _remoteStreams.add(stream);
        };
        await _peerConnection?.addStream(_localStream!);
        break;
      case Constants.unifiedPlan:

        /// get remote stream ///
        _peerConnection?.onTrack = (event) {
          if (event.track.kind == 'video') {
            onAddRemoteStream.call(event.streams[0]);
          }
        };

        /// get local stream ///
        _localStream?.getTracks().forEach((track) async {
          if(_peerConnection!=null) {
            _senders.add(await _peerConnection!.addTrack(track, _localStream!));
          }
        });

        break;
    }

    /// remove remote stream ///
    _peerConnection?.onRemoveStream = (stream) {
      onRemoveRemoteStream?.call(stream);
      _remoteStreams.removeWhere((it) {
        return (it.id == stream.id);
      });
    };
  }


  /// Create Local Stream ///
  @override
  Future<MediaStream> createStream() async {
    final Map<String, dynamic> mediaConstraints = {
      'audio': true,
      'video': {
        'mandatory': {
          'minWidth': '320',
          'minHeight': '240',
          'minFrameRate': '30',
        },
        'facingMode': 'user',
        'optional': [],
      }
    };
    MediaStream stream = await navigator.mediaDevices.getUserMedia(mediaConstraints);
    return stream;
  }

  /// get peerconnection instance ///
  @override
  RTCPeerConnection? getPeerConnection() {
    return _peerConnection;
  }


  /// stop video call ///
  @override
  Future<void> stop() async {
    final param = {
      'id':Constants.stop
    };
    await getSocket().send(const JsonEncoder().convert(param));
  }

  @override
  Future<void> publishCallVideo({
    required Map<String,dynamic> from,
    required Map<String,dynamic> to,
    required RTCSessionDescription sessionDescription}) async {

    final param = {
      'id':Constants.call,
      'from':from,
      'to':to,
      'sdpOffer':sessionDescription.sdp
    };

    await getSocket().send(const JsonEncoder().convert(param));
  }

  /// publish video call ///
  @override
  Future<void> publishIncomingCallVideo({
    required Map<String,dynamic> to,
    required RTCSessionDescription sessionDescription,
    Map<String, dynamic>? consultation
  }) async {
    final param = {
      'id': Constants.incomingCallResponse,
      'from': to,
      'callResponse': Constants.accept,
      'consultationPayload': consultation,
      'sdpOffer': sessionDescription.sdp
    };
    await getSocket().send(const JsonEncoder().convert(param));
  }

  /// on ice candidate recive from socket ///
  @override
  Future<void> iceCandidate(Map<String, dynamic> candidate) async {
    RTCIceCandidate iceCandidate = RTCIceCandidate(candidate['candidate'], candidate['sdpMid'], candidate['sdpMLineIndex']);
    RTCSessionDescription? remoteDescription = await _peerConnection?.getRemoteDescription();

    switch(_peerConnection?.signalingState){
      case RTCSignalingState.RTCSignalingStateClosed:
        debugPrint("saveIceCandidate error: PeerConnection object is closed");
        break;
      case RTCSignalingState.RTCSignalingStateStable:
        if(remoteDescription!=null){
          _peerConnection?.addCandidate(iceCandidate);
        }else{
          remoteCandidates.add(iceCandidate);
        }
        break;
      default:
        remoteCandidates.add(iceCandidate);
    }
  }


  /// receive video answer web rtc ///
  @override
  Future<void> receiveVideoAnswer(String sdpAnswer) async {
    RTCSessionDescription sessionDescription = RTCSessionDescription(sdpAnswer,'answer');
    _peerConnection?.setRemoteDescription(sessionDescription);

    /// add candidate to local peer connection
    if (remoteCandidates.isNotEmpty) {
      for (var candidate in remoteCandidates) {
        await _peerConnection?.addCandidate(candidate);
      }
      remoteCandidates.clear();
    }
  }

  /// send ice Candidate to web socket ///
  @override
  Future<void> onIceCandidate(RTCIceCandidate rtcIceCandidate) async {

    ///set ice candidate from local peerconnection and send to websocket///
    final candidate = {
      'sdpMid': rtcIceCandidate.sdpMid,
      'sdpMLineIndex': rtcIceCandidate.sdpMLineIndex,
      'candidate': rtcIceCandidate.candidate
    };

    final param = {
      'id': Constants.onIceCandidate,
      'candidate' : candidate
    };
    await getSocket().send(const JsonEncoder().convert(param));
  }

  /// create offer webrtc ///
  @override
  Future<void> createOffer({
    required Map<String, dynamic> from,
    required Map<String, dynamic> to,
    required bool isCalling,
    Map<String, dynamic>? consultation
  }) async {
    try {
      RTCSessionDescription sessionDescription = await _peerConnection!.createOffer(Constants.sdpConstraint);
      await _peerConnection?.setLocalDescription(sessionDescription);
      if(isCalling){
        publishCallVideo(from: from, to: to, sessionDescription: sessionDescription);
      }else {
        publishIncomingCallVideo(to: to, sessionDescription: sessionDescription, consultation: consultation);
      }
    } catch (e) {
      debugPrint(e.toString());
    }
  }

  /// close session ///
  @override
  Future<void> closeSession() async {
    _localStream?.getTracks().forEach((element) async {
      await element.stop();
    });
    await _localStream?.dispose();
    _localStream = null;

    _senders.clear();
    await _peerConnection?.close();
    _peerConnection = null;
  }


  /// send message to socket if calle registered to web socket for action call from caller
  @override
  Future<void> notifyCaller({required Map<String, dynamic> from, required String to}) async{
    final registrationSessionTo = {
      'name': to
    };

    final param = {
      'id': Constants.notifyCaller,
      'from': from,
      'to': registrationSessionTo
    };
    await getSocket().send(const JsonEncoder().convert(param));
  }


  /// reject video call///
  @override
  Future<void> reject({required Map<String,dynamic> from, required Map<String,dynamic> to}) async{
    final param = {
      'id': Constants.rejectCalling,
      'from': from,
      'to': to
    };
    await getSocket().send(const JsonEncoder().convert(param));
  }

  /// enable disable mic in video call ///
  @override
  void muteMic() {
    if (_localStream != null) {
      bool enabled = _localStream!.getAudioTracks()[0].enabled;
      _localStream!.getAudioTracks()[0].enabled = !enabled;
    }
  }

  /// switch camera front or rear ///
  @override
  void switchCamera({required VideoSource videoSource, required Function(MediaStream stream) onLocalStream}) {
    if (_localStream != null) {
      if (videoSource != VideoSource.camera) {
        for (var sender in _senders) {
          if (sender.track!.kind == 'video') {
            sender.replaceTrack(_localStream!.getVideoTracks()[0]);
          }
        }
        onLocalStream.call(_localStream!);
      } else {
        Helper.switchCamera(_localStream!.getVideoTracks()[0]);
      }
    }
  }

  /// enable disable video ///
  @override
  void enableVideo() {
    if (_localStream != null) {
      bool enabled = _localStream!.getVideoTracks()[0].enabled;
      _localStream!.getVideoTracks()[0].enabled = !enabled;
    }
  }

  @override
  Future<void> callingMobileDevice({required String from, required String to, required Map<String, dynamic>? fcm}) async {
    final param = {
      'id': Constants.calling,
      'from': from,
      'to': to,
      'fcm': fcm
    };
    await getSocket().send(const JsonEncoder().convert(param));
  }

  /// switch to screen share
  void switchToScreenSharing(MediaStream stream, VideoSource videoSource, Function(MediaStream stream) onLocalStream) {
    if (_localStream != null && videoSource == VideoSource.screen) {
      for (var sender in _senders) {
        if (sender.track!.kind == 'video') {
          sender.replaceTrack(stream.getVideoTracks()[0]);
        }
      }
      onLocalStream.call(stream);
    }
  }

  /// select screen share
  @override
  Future<void> selectScreenSource({required VideoSource videoSource, required Function(MediaStream stream) onLocalStream}) async {
    MediaStream? screenStream;
    if (WebRTC.platformIsWeb) {

      final Map<String, dynamic> mediaConstraints = {
        'audio': false,
        'video': true,
      };

      screenStream = await navigator.mediaDevices.getDisplayMedia(mediaConstraints);
    }
    if (screenStream != null) switchToScreenSharing(screenStream, videoSource, onLocalStream);
  }
}